#include <ntddk.h>
#include "hookmon.h"

#define DEBUG

//   
UNICODE_STRING	DeviceName;
UNICODE_STRING	SymbolicLinkName;
PDEVICE_OBJECT	deviceObject = NULL;
PUSHORT		aatomSysLoaded = NULL;
ULONG			ClientIdOffset = 0;

PVOID
GetInfoTable (
	IN ULONG ATableType)
{
	ULONG		mSize = 0x8000;
	PVOID		mPtr;
	NTSTATUS	status;
	do {
		mPtr = ExAllocatePool(PagedPool, mSize);

		if (!mPtr) return NULL;

		memset(mPtr, 0, mSize);

		status = ZwQuerySystemInformation(ATableType, mPtr, mSize, NULL); 
		
		if (status == STATUS_INFO_LENGTH_MISMATCH) {
			ExFreePool(mPtr);
			mSize = mSize * 2;
			}

		} while (status == STATUS_INFO_LENGTH_MISMATCH);

	if (NT_SUCCESS(status)) return mPtr;

	ExFreePool(mPtr);

	return NULL;
}

PVOID
GetWin32kBase()
{
	PVOID				result = NULL;
	ULONG             	i;
	PSYSTEM_MODULE_INFORMATION_EX	smi = GetInfoTable(SystemModuleInformation);

	if (smi) {
		for( i = 0; i < smi->ModulesCount; i++ ) {
			if (!_stricmp("win32k.sys", &smi->Modules[i].ImageName[smi->Modules[i].ModuleNameOffset]))
				{
				result = smi->Modules[i].Base;
				break;
				}
			}

		ExFreePool(smi);
		}

	return result;
}

/*
	GetaatomSysLoadedAddress()
	Should be only executed in GUI-process' context
*/
ULONG
GetaatomSysLoadedAddress()
{
	ULONG		i, result = 0;
	PVOID		Win32kBase = GetWin32kBase();

	if (Win32kBase) {
		for (i = (ULONG)Win32kBase + 0x1000; i < (ULONG)Win32kBase + 0x000FFFFF; i++) {
			if ( (*(ULONG*)i == 0x753C3966) && 		// cmp	aatomSysLoaded[esi*2], di
			     ( (*(SHORT*)(i + 8) == 0x0574) || 	// jz		+05
			     (*(SHORT*)(i + 8) == 0x1C75) ) ) {	// jnz 	+1C	[win32k 5.0.2195.6708]
				result = *(ULONG*)(i + 4);
				break;
				}
			}
		}

	return result;
}

ULONG
GetClientIdOffset()
{
	PCWSTR		pFuncName       = L"PsGetCurrentThreadId";
	UNICODE_STRING	usFuncName;
	PVOID			pFuncAddr;
	ULONG			result = 0;

	RtlInitUnicodeString(&usFuncName, pFuncName);
	pFuncAddr = MmGetSystemRoutineAddress(&usFuncName);
	if (pFuncAddr) {
		result = *(PULONG)((ULONG)pFuncAddr + 8);
		}

	return	result - 4;
}

NTSTATUS
DriverDispatcher(
		IN PDEVICE_OBJECT DeviceObject,
		IN PIRP Irp)
{
	PIO_STACK_LOCATION 		irpStack;
	ULONG 				InBuffSize, OutBuffSize;
	PVOID 				pBuff;
	PMEMORY_REQUEST			mr;
	WCHAR					libPath[256];
	PCLIENT_ID				pCid, pResultCid;
	NTSTATUS				status;

	irpStack = IoGetCurrentIrpStackLocation (Irp);
	InBuffSize   = irpStack->Parameters.DeviceIoControl.InputBufferLength;
	pBuff = Irp->AssociatedIrp.SystemBuffer;
	Irp->IoStatus.Information = 0;

	switch (irpStack->MajorFunction) {
		case IRP_MJ_DEVICE_CONTROL:
			switch (irpStack->Parameters.DeviceIoControl.IoControlCode) {
				case IOCTL_READ_KERNEL_MEMORY:
					if (InBuffSize >= sizeof(MEMORY_REQUEST)) {
						mr = pBuff;
						__try {
							ProbeForWrite(mr->pDst, mr->dwBytesCount, 1);
							memcpy(mr->pDst, mr->pSrc, mr->dwBytesCount);
							status = STATUS_SUCCESS;
							}
						__except (EXCEPTION_EXECUTE_HANDLER) {
							status = GetExceptionCode();
							}
						}
					else {
						status = STATUS_BUFFER_TOO_SMALL;
						}
					break;
				case IOCTL_GET_MODULE_ATOM:
					if (!aatomSysLoaded) {
						aatomSysLoaded = (PVOID)GetaatomSysLoadedAddress();
						}

					if ( (InBuffSize >= sizeof(ULONG)) && aatomSysLoaded ) {
						*(ULONG*)pBuff = aatomSysLoaded[*(ULONG*)pBuff];
						Irp->IoStatus.Information = sizeof(ULONG);
						status = STATUS_SUCCESS;
						}
					else {
						status = STATUS_BUFFER_TOO_SMALL;
						}
					break;
				case IOCTL_GET_CID_BY_PTI:
					if (!ClientIdOffset) {
						ClientIdOffset = GetClientIdOffset();
						}

					if ( (InBuffSize >= sizeof(CLIENT_ID)) && ClientIdOffset ) {
						pResultCid = (PCLIENT_ID)pBuff;
						pCid = (PCLIENT_ID)(*(ULONG*)(*(ULONG*)pBuff) + ClientIdOffset);
						pResultCid->UniqueProcess = pCid->UniqueProcess;
						pResultCid->UniqueThread = pCid->UniqueThread;
						Irp->IoStatus.Information = sizeof(CLIENT_ID);
						status = STATUS_SUCCESS;
						}
					else {
						status = STATUS_BUFFER_TOO_SMALL;
						}
					break;
				default:
					status = STATUS_INVALID_DEVICE_REQUEST;
				}
			break;
		default:
			status = STATUS_SUCCESS;	
		}
		
	Irp->IoStatus.Status = status;
	IoCompleteRequest(Irp, IO_NO_INCREMENT);

	return STATUS_SUCCESS;
}


#ifdef DEBUG
VOID
DriverUnload(
	IN PDRIVER_OBJECT DriverObject)
{
	IoDeleteSymbolicLink(&SymbolicLinkName);
	IoDeleteDevice(deviceObject);

	return;
}
#endif

NTSTATUS
DriverEntry(
		IN PDRIVER_OBJECT DriverObject,
		IN PUNICODE_STRING RegistryPath)
{
	NTSTATUS	st;
	PCWSTR		dDeviceName       = L"\\Device\\hookmon";
	PCWSTR		dSymbolicLinkName = L"\\DosDevices\\hookmon";

	RtlInitUnicodeString(&DeviceName, dDeviceName);
	RtlInitUnicodeString(&SymbolicLinkName, dSymbolicLinkName);

	st = IoCreateDevice(DriverObject, 0, &DeviceName, FILE_DEVICE_NULL, 0, FALSE, &deviceObject);
	if (NT_SUCCESS(st)) {
		st = IoCreateSymbolicLink(&SymbolicLinkName, &DeviceName);
		if (NT_SUCCESS(st)) {
			DriverObject->MajorFunction[IRP_MJ_CREATE] =
			DriverObject->MajorFunction[IRP_MJ_CLOSE]  = 
			DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = DriverDispatcher;
#ifdef DEBUG
			DriverObject->DriverUnload = DriverUnload;
#endif
			}
		else {
			IoDeleteDevice(deviceObject);
			}
		}
	
	return st;
}