// Coded by Twister [http://twister.rootkits.ru]
// Last update 12.10.2008

unit winuser;

interface

const
  TYPE_FREE           = 0;        // must be zero!
  TYPE_WINDOW         = 1;        // in order of use for C code lookups
  TYPE_MENU           = 2;
  TYPE_CURSOR         = 3;
  TYPE_SETWINDOWPOS   = 4;
  TYPE_HOOK           = 5;
  TYPE_CLIPDATA       = 6;        // clipboard data
  TYPE_CALLPROC       = 7;
  TYPE_ACCELTABLE     = 8;
  TYPE_DDEACCESS      = 9;
  TYPE_DDECONV        = 10;
  TYPE_DDEXACT        = 11;       // DDE transaction tracking info.
  TYPE_MONITOR        = 12;
  TYPE_KBDLAYOUT      = 13;       // Keyboard Layout handle (HKL) object.
  TYPE_KBDFILE        = 14;       // Keyboard Layout file object.
  TYPE_WINEVENTHOOK   = 15;       // WinEvent hook (EVENTHOOK)
  TYPE_TIMER          = 16;
  TYPE_INPUTCONTEXT   = 17;       // Input Context info structure
  TYPE_CTYPES         = 18;       // Count of TYPEs; Must be LAST + 1
  TYPE_GENERIC        = 255;      // used for generic handle validation

  HF_GLOBAL           = $0001;
  HF_ANSI             = $0002;
  HF_NEEDHC_SKIP      = $0004;
  HF_HUNG             = $0008;      // Hook Proc hung don't call if system
  HF_HOOKFAULTED      = $0010;      // Hook Proc faulted
  HF_NOPLAYBACKDELAY  = $0020;      // Ignore requested delay
  HF_WX86KNOWNDLL     = $0040;      // Hook Module is x86 machine type
  HF_DESTROYED        = $0080;      // Set by FreeHook
  HF_INCHECKWHF       = $0100;      // fsHooks is being updated
  HF_FREED            = $0200;      // Object has been freed.
  HF_DBGUSED          = $03FF;      // Update if adding a flag

  HANDLEF_DESTROY        = $01;
  HANDLEF_INDESTROY      = $02;
  HANDLEF_INWAITFORDEATH = $04;
  HANDLEF_FINALDESTROY   = $08;
  HANDLEF_MARKED_OK      = $10;
  HANDLEF_GRANTED        = $20;
  HANDLEF_VALID          = $3F;

  HMINDEXBITS = $0000FFFF;
  HMUNIQSHIFT = 16;
  HMUNIQBITS  = $FFFF0000;

type
  DWORD = longint;

  PHEAD = ^HEAD;
    HEAD = packed record
      h: DWORD;                       //  
      cLockObj: DWORD;
      end;

  PTHROBJHEAD = ^THROBJHEAD;
    THROBJHEAD = packed record
      head: HEAD;
      pti: pointer;                   // PTHREADINFO
      end;

  PDESKHEAD = ^DESKHEAD;
    DESKHEAD = packed record
      rpdesk: pointer;                // PDESKTOP
      pSelf: pointer;
      end;

  PTHRDESKHEAD = ^THRDESKHEAD;
    THRDESKHEAD = packed record
      throbjhead: THROBJHEAD;
      deskhead: DESKHEAD;
      end;

  PHOOK = ^HOOK;
    tagHOOK = packed record
      head: THRDESKHEAD;
      phkNext: PHOOK;
      iHook: integer;		              // WH_xxx hook type
      offPfn: DWORD;		              // ((ULONG_PTR)pfnFilterProc) - ((ULONG_PTR)hmod)
      flags: DWORD;		                // HF_xxx flags
      ihmod: integer;		              // .       GetClipboardFormatName()
      ptiHooked: pointer;	            // Thread hooked.
      rpdesk: pointer;		            // Global hook pdesk. Only used when
			  			                        //  hook is locked and owner is destroyed
      end;
    HOOK = tagHOOK;

  PHANDLEENTRY = ^HANDLEENTRY;
    HANDLEENTRY = packed record
      pHead: pointer;                 //    (pWnd   )
      pOwner: pointer;                //     (ppi  pti)
      bType: BYTE;                    //   (TYPE_xxx)
      bFlags: BYTE;                   // 
      wUniq: WORD;                    // uniqueness count
      end;
  PHANDLEENTRY_ARRAY = ^HANDLEENTRY_ARRAY;
    HANDLEENTRY_ARRAY = array[0..0] of HANDLEENTRY;

  PSHAREDINFO = ^SHAREDINFO;
    SHAREDINFO = packed record
      psi: pointer;                   //    SERVERINFO
      aheList: PHANDLEENTRY_ARRAY;    //    
      pDispInfo: pointer;             //    DISPLAYINFO
      ulSharedDelta: DWORD;           //    
                                      //    
                                      //    .
      end;

  function HandleFromIndex(Index: integer): DWORD;
  function IndexFromHandle(hWnd: DWORD): integer;
  function UniqFromHandle(hWnd: DWORD): WORD;
  function ValidateHwnd(hWnd: DWORD): pointer;
  function GetKernelPHE(hWnd: DWORD): pointer;
  function GetPClassName(pWnd: pointer): pointer; stdcall;
  function GetSharedInfoAddress(): PSHAREDINFO;

var
  gSharedInfo: PSHAREDINFO;

implementation

uses
  windows;

function HandleFromIndex(Index: integer): DWORD;
begin
result := Index or (gSharedInfo.aheList[Index].wUniq shl HMUNIQSHIFT);
end;

function IndexFromHandle(hWnd: DWORD): integer;
begin
result := hWnd and HMINDEXBITS;
end;

function UniqFromHandle(hWnd: DWORD): WORD;
begin
result := (hWnd shr HMUNIQSHIFT) and HMUNIQBITS;
end;

function ValidateHwnd(hWnd: DWORD): pointer;
var
  phe: PHANDLEENTRY;
  dwDelta: DWORD;
begin
{
  Vista SP1    phe.wUniq = UniqFromHandle(hWnd).
     ...
}
result := nil;
asm
  mov   eax, fs:[$18]             //     TEB    ,
  lea   eax, [eax + $6CC]         //      
  mov   eax, dword [eax + $1C]    //   win_2k   win_vista
  mov   dwDelta, eax
end;
phe := @gSharedInfo.aheList[IndexFromHandle(hWnd)];
if phe.bType = TYPE_WINDOW then
  result := pointer(DWORD(phe.pHead) - dwDelta);
end;

function GetKernelPHE(hWnd: DWORD): pointer;
var
  phe: PHANDLEENTRY;
begin
phe := @gSharedInfo.aheList[IndexFromHandle(hWnd)];
result := pointer(DWORD(phe) + gSharedInfo.ulSharedDelta)
end;

function GetPClassName(pWnd: pointer): pointer; stdcall;
asm
  pushad
  mov   eax, pWnd
  mov   edx, dword [eax + $10] // tagWND.head.pSelf (kernel addr)
  mov   ecx, dword [eax + $64] // tagWND.pcls (kernel addr)
  sub   ecx, edx
  add   ecx, eax               // tagWND.pcls (user addr)
  mov   esi, dword [ecx + $54] // tagWND.pcls.lpszClientAnsiMenuName (kernel addr)
  sub   esi, edx
  add   esi, eax               // tagWND.pcls.lpszClientAnsiMenuName (user addr)
  mov   result, esi
  popad
end;

function GetSharedInfoAddress(): PSHAREDINFO;
var
  WowHandlers: pointer;
  i: DWORD;
begin
result := nil;
WowHandlers := GetProcAddress(GetModuleHandle('user32.dll'), 'UserRegisterWowHandlers');
if WowHandlers <> nil then
  for i := DWORD(WowHandlers) to DWORD(WowHandlers) + $1000 do
    if (WORD(pointer(i)^) = WORD($40C7)) and (BYTE(pointer(i + 7)^) = BYTE($B8)) then
      begin
      result := pointer(DWORD(pointer(i + 8)^));
      exit;
      end;
end;

begin
  gSharedInfo := GetSharedInfoAddress();
end.
